#!/usr/bin/php
<?
/* filterctl controls the filterwheel inside the dewar for the COAST IR camera. It does some simple I/O	*
 * using ftdi245io, in order to read the current position of the filter wheel, and allow it to be	*
 * changed. The ftdi245 USB device is connected to the digital interface PCB and to the copper-box via  *
 * optical fibre. This can read the position  of the filters, or can drive a stepper motor to change it.*
 * Some steps are positioned between filters, rather than on a filter itself. The interface between the *
 * USB-245 module and the filter control pcb is detailed below; it is an 8-bit I/O port, where 2 bits  	*
 * are outputs (dewar_power, stepper), and 3 bits are inputs: 3 light-gates, which read 6 possible      *
 * positions, or "in-between" (or "invalid"). When the dewar is powered off, the 3 input channels are   *
 * re-purposed to allow monitoring of the LiPo battery voltage levels. The last 3 channels of the       *
 * FTDI245 are used to control clocking of the pulseblaster: how should the 5kHz master be divided.     *
 * See also $USAGE below. Download from www.richardneill.org/source.php . Feedback is always welcome.	*

 * Copyright (C) Richard Neill 2010-2012, <filterctl at REMOVE.ME.richardneill.org>. This program is Free Software. You can
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html */


#Idea: could we be really clever and distinguish a totally dead battery from a flat (but not quite dead) one, by checking what the light gates read?
#      i.e. a light gate reading should normally not be all zeros.
#     => Not in the current config: all zeros means misaligned, which can happen already. If all zeros were the illegal setting, then we could.


$VERSION="2.1";
$RELEASE_DATE="2012-04-04";
$AUTHOR="Richard Neill";
$EMAIL="<filterctl at richardneill.org>";
$COPYRIGHT_DATES="2010-2012";
$WEB="http://www.richardneill.org/source.php";
$LICENSE="This is Free Software, licensed under the GNU GPL, v.3 or later.";

//--------------------------------------------------------------------------------------------------------------
// **** BEGIN CONFIGURATION SECTION: ****
$DEBUG=false;					//Debugging on?  (Can also be enabled with -d)
$DUMMY=false;					//Dummy mode? This uses a fake version of FTDI245IO_BIN, in order to test without the USB module being present.

$FTDI245IO_BIN="ftdi245io";			//The program used to to the I/O for the filter control circuit.
$FTDI245IO_WRITE="--porta_setpins";		//The port/argument used by ftdio245 to write data. (Corresponds to a particular FTDI USB245 module).
$FTDI245IO_READBACK="--porta_readback";		//The port/argument used by ftdio245 to read back the buffered byte (inputs + last outputs, depending on mask). Returns a single byte in ASCII form, eg "0x0C"
$DUMMY_VALUE = 0x01;				//In dummy mode: return this byte value. (used for testing).

$IO_MASK=0xF8;					//The output mask on this port. 0s are inputs, 1s are outputs.  Used as sanity check. Must be same as ftdi245io.h -> PORTA_OUTPUT_MASK
$LIGHT_GATE_BITS=0x07;				//The bits which correspond to the 3 light gates. (Bits 2,1,0 are the three sensors; 8 permutations). Also used for Battery readings.
$STEP_BIT=0x08;					//Falling edge on this bit makes the wheel step once.  (Bit 3)
$POWER_BIT=0x10;				//The bit which controls power. (High => Power on the dewar (enable filter step, and leds); Low => power down the dewar. (battery-readout mode))
$CLOCK_DIVIDER_BITS=0xE0;			//The bits used for the clock frequency divider: (Bits 7,6,5)

//Array of filter mappings. This array has 4 sections: bits (the bit-map for the light-gate codes), key (the short filtername, used by -f, no spaces), desc (the filter description text), sequence (the sequence in the physical device).
$FILTERS[0]['bits'] = 0x00;
$FILTERS[0]['key']  = "Misaligned";		//See also: $FILTER_MISALIGNED.
$FILTERS[0]['desc'] = "The filter-wheel is not aligned with any filter, but part-way between.";
$FILTERS[0]['seq']  = -1;			//-1 for invalid sequence.

$FILTERS[1]['bits'] = 0x1;			//Black anodised solid metal disc.
$FILTERS[1]['key']  = "Black";			
$FILTERS[1]['desc'] = "F1. Solid metal disc"; 
$FILTERS[1]['seq']  = 1;			

$FILTERS[2]['bits'] = 0x5;			//Deep purple colour, labelled as "1036.5/18.0 NM 77 K 4197 BARR +3".
$FILTERS[2]['key']  = "1036nm";			//Fractionally too wide. Too thick (space is 5mm, height is 6.5mm), so held in as interference fit.
$FILTERS[2]['desc'] = "F2. 1036.5 / 18 nm. (Appears Purple)";
$FILTERS[2]['seq']  = 2;

$FILTERS[3]['bits'] = 0x4;
$FILTERS[3]['key']  = "Open";			//Open hole (25mm diameter)
$FILTERS[3]['desc'] = "F3. Open";
$FILTERS[3]['seq']  = 3;

$FILTERS[4]['bits'] = 0x6;
$FILTERS[4]['key']  = "2205nm";			//2205 nm, half-power bandwidth 150 nm. Appears gold on top (away from sensor) and bright blue on bottom.
$FILTERS[4]['desc'] = "F4. 2205 / 150 nm. (Appears Gold)";
$FILTERS[4]['seq']  = 4;

$FILTERS[5]['bits'] = 0x2;
$FILTERS[5]['key']  = "1287nm";			//1287 nm, half-power bandwidth 35 nm. Appears green on top (away from sensor) and slightly darker green on bottom.
$FILTERS[5]['desc'] = "F5. 1287 / 35 nm. (Appears Green)";
$FILTERS[5]['seq']  = 5;

$FILTERS[6]['bits'] = 0x3;
$FILTERS[6]['key']  = "Quad_1mm";		//Solid Aluminium (not anodised), 4 holes (not aligned rotationally)
$FILTERS[6]['desc'] = "F6. 4 x 1mm holes, metal disc.";
$FILTERS[6]['seq']  = 6;

$FILTERS[7]['bits'] = 0x7;
$FILTERS[7]['key']  = "Illegal";		//See also $FILTER_ILLEGAL.
$FILTERS[7]['desc'] = "Error: All 3 light-gates simultaneously open. Check correctly assembled, filter wheel in place?";
$FILTERS[7]['seq']  = -1;

$FILTER_MISALIGNED=0;				//These are the special-case filter "positions" that cannot be a sane target of -g
$FILTER_ILLEGAL=7;

$STEPPER_STEPS = 24;				//Number of steps of the motor per rev.

//Array of voltage mappings. This is when the dewar is powered off, and the inputs are used for battery levels instead. Unlike the filters, this isn't really multplexed:
//there are just 3 comparators, and it acts like a bargraph. So only the following values are legal: 000   001  011  111. These values are actually adjustable using pots within the box.
//See also http://www.ibt-power.com/Battery_packs/Li_Polymer/Lithium_polymer_tech.html and lipo.txt
$VOLTAGES[0]['bits'] = 0x00;			//3 fibres lit, 111,  inverts to 000.
$VOLTAGES[0]['key']  = "Full";
$VOLTAGES[0]['desc'] = "> 11.5V, 80%+ charge";

$VOLTAGES[1]['bits'] = 0x01;
$VOLTAGES[1]['key']  = "Illegal";
$VOLTAGES[1]['desc'] = "Invalid";

$VOLTAGES[2]['bits'] = 0x02;
$VOLTAGES[2]['key']  = "Illegal";
$VOLTAGES[2]['desc'] = "Invalid";

$VOLTAGES[3]['bits'] = 0x03;
$VOLTAGES[3]['key']  = "Illegal";
$VOLTAGES[3]['desc'] = "Invalid";

$VOLTAGES[4]['bits'] = 0x04;			//2 fibres lit, 011,  inverts to 100
$VOLTAGES[4]['key']  = "Mid";
$VOLTAGES[4]['desc'] = "11.0V - 11.5V, 40-80% charge";

$VOLTAGES[5]['bits'] = 0x05;
$VOLTAGES[5]['key']  = "Illegal";
$VOLTAGES[5]['desc'] = "Invalid";

$VOLTAGES[6]['bits'] = 0x06;			//1 fibre lit, 001, inverts to 110
$VOLTAGES[6]['key']  = "Low";
$VOLTAGES[6]['desc'] = "10.0V - 11.0V, 5-40% charge";

$VOLTAGES[7]['bits'] = 0x07;			//no fibres lit, 000, inverts to 111.
$VOLTAGES[7]['key']  = "Flat";
$VOLTAGES[7]['desc'] = "< 10.0V, < 5% charge";

$VOLTAGE_SAFETY_CUTOUT = "9.0V";		//The voltage at which the hardware cutout operates.

$VOLTAGES_KEYS_ILLEGAL = array (1,2,3,5);  	//The keys of $VOLTAGE which are invalid/illegal.
$VOLTAGES_KEY_FLAT=7;				//The key of $VOLTAGE which has the special meaning of "flat".


//Array of clock bit mappings. These are for the outputs for the clock divider.
$CLOCKS[0]['bits'] = 0x00;
$CLOCKS[0]['key']  = "Ext";
$CLOCKS[0]['desc'] = "External, Manual or Software triggering";

$CLOCKS[1]['bits'] = 0x20;
$CLOCKS[1]['key']  = "50";
$CLOCKS[1]['desc'] = "50 Hz (= 5kHz/100, opt:125 Hz)";

$CLOCKS[2]['bits'] = 0x40;
$CLOCKS[2]['key']  = "100";
$CLOCKS[2]['desc'] = "100 Hz (= 5kHz/50)";

$CLOCKS[3]['bits'] = 0x60;
$CLOCKS[3]['key']  = "250";
$CLOCKS[3]['desc'] = "250 Hz (= 5kHz/20, opt:200 Hz)";

$CLOCKS[4]['bits'] = 0x80;
$CLOCKS[4]['key']  = "500";
$CLOCKS[4]['desc'] = "500 Hz (= 5kHz/10)";

$CLOCKS[5]['bits'] = 0xA0;
$CLOCKS[5]['key']  = "1000";
$CLOCKS[5]['desc'] = "1 kHz (= 5kHz/5)";

$CLOCKS[6]['bits'] = 0xC0;
$CLOCKS[6]['key']  = "2500";
$CLOCKS[6]['desc'] = "2.5 kHz (= 5kHz/2)";

$CLOCKS[7]['bits'] = 0xE0;
$CLOCKS[7]['key']  = "5000";
$CLOCKS[7]['desc'] = "5 kHz (= 5kHz/1)";

$POWER_BIT_ACTIVE      = 'LOW';		//"HIGH" or "LOW".  Power bit is active low.
$STEP_BIT_ACTIVE_GOING = 'LOW';		//"HIGH" or "LOW".  Step bit is active on falling edges (i.e. going low). (Loopback is enabled when step is active and power is off)

$INTERVAL_POSTPOWER_MS=500;		//Time (ms) to wait after changing the power state of the dewar electronics (to allow it to stabilise) before we do anything. (100ms is too short).
$INTERVAL_PREPOWER_NOGLITCH_MS=100;	//Time (ms) to hold the motor step inactive, before we turn on power. This prevents the motor from jumping when we just want to measure it.
$INTERVAL_STEP_MS=3000;			//Time (ms) to wait for the 555 chip to pulse (by dead-reckoning; must exceed the RC-time-constant). Also allow the filterwheel oscillation to damp (after the pulse).
$INTERVAL_ALIGNMENTCHECK_MS=200;	//Time (ms) to wait between polling during alignment checks.
$INTERVAL_BATTERY_POLL_S=60; 		//Time (s) to wait between battery polling in monitor mode. (NB defined as 1 "minute" also)
$INTERVAL_BATTERY_POLLFAST_S=1;		//Time (s) to wait between battery polling in fast monitor mode.

$MAX_STEP_ATTEMPTS=72;		//Max number of steps we attempt before giving up. (At 24 steps/rev, this is 3 revolutions).
$SELFTEST_STEPS=360;		//Max number of steps in a selftest. Don't want to wear out the motor. (NB it has no real lubrication, since it is designed for Liquid nitrogen and vacuum).
$SELFTEST_MAX_MISALIGNED=($STEPPER_STEPS/4);	//Max number of consecutive steps during a selftest which can all be misaligned, before a warning is given.

// **** END CONFIGURATION SECTION ****

if ($DUMMY){	//In dummy mode, fake the control circuit.
	$dummy_hex_str = sprintf("0x%02x", $DUMMY_VALUE);	//Return this byte string (format as "0xXX"). The dummy command, used for testing must return a single byte, and ignore its args.
	$DUMMY_BIN="echo $dummy_hex_str #dummy_cmd";		//So use the comment trick. Thus "echo 0xff #dummy_cmd --porta_setpins 0xfe" just becomes "echo 0xff"
	print_msg("Warning: using dummy interface: '$DUMMY_BIN' instead of '$FTDI245IO_BIN'.\n");
	$FTDI245IO_BIN=$DUMMY_BIN;
}

//--------------------------------------------------------------------------------------------------------------
//ERROR REPORTING, and PHP CONFIGURATION:

$display_errors=ini_get('display_errors');	//Display errors ought to be on. Otherwise, if this program has a parse error, php will just unhelpfully exit silently!  It can be enabled in /etc/php-cli.ini.
if (!$display_errors){				//Note: we can't just ini_set('display_errors', 'on') because the parse-error will get to us first!
	debug_print_msg("Warning: 'display_errors' is set to OFF in the php.ini file. This will cause this script to die silently if there are any parse errors in it. ".
			"Note: 'display_errors' *should* be off for a webserver (in '/etc/php.ini'), but can be turned on for PHP-CLI (in '/etc/php5/cli/php.ini').");
}
if (php_sapi_name()!= "cli"){			//Check we're using the PHP-CLI sapi (should always be true, but someone might try to run this in a webserver!)
	fatal_error("for some reason, this is not using the PHP-CLI SAPI, but the '".php_sapi_name()."' SAPI. Please use PHP-CLI.");
}
if (substr(phpversion(),0,1) < 5){  //Test PHP Version. However, this is probably unnecessary: it ought to work with *any* recent PHP (after 2005 ish)
	fatal_error("this requires at least version 5.x of PHP. However, the installed version is only version '$phpversion'.");
}
error_reporting (E_ALL & ~ E_NOTICE);	//Turn on all error reporting and warnings, except for Notices.

//--------------------------------------------------------------------------------------------------------------
//USAGE:

$binary_name=basename($argv[0]);  //DEVEL = "filterctl.php"; INSTALLED ="filterctl"
$USAGE = <<<EOT
USAGE:
	$binary_name [ OPTIONS ]

DESCRIPTION:
	$binary_name interfaces with a control circuit (via a USB-FTDI245 device) to communicate
	with some light-gates and a stepper motor inside the dewar. The light gates detect which
	position the filter-wheel is in; the stepper motor can be stepped to change this. The
	entire circuit is usually turned off in order to eliminate heat-dissipation and 'glow'
	inside the dewar; this is done automatically if the program exits gracefully.
	$binary_name can also configure the clock-divider, monitor the battery, and enable loopback.

HARDWARE:
	The dewar contains a filter wheel containing 6 filters, mounted upon the spindle of a
	stepper-motor: this motor always comes to rest in one of 24 positions, which include the
	6 correct ones. Position readout is done by 3 light-gates, (formed by red LEDs and
	photo-transistors) which can distinguish the 6 correct positions, the misaligned-state,
	or the impossible-state. In order to avoid waste-heat and stray-light, $binary_name powers
	down the stepper-motor, the light-gates, and the 10uA temp-sensor after it has exited.

	When the dewar internals are powered-down, the optical fibres are re-purposed to carry
	battery voltage information. There are 3 bits, carrying 4 possible voltage states (the
	CMOS gate-comparator circuit is simple and cannot use all 8 permutations).

	In Loopback mode, two of the 4 channels are re-purposed: SPDT relays swap from the
	Hawaii quadrants (3,2) to digital loopback (Line3) and dewar temperature monitor.
	(The first two quadrants are always connected to the ADC).
	NORMAL:
	      ai3+   Hawaii 4th quadrant source follower, 10k pullup to V_A+5.
	      ai2+   Hawaii 3rd quadrant source follower, 10k pullup to V_A+5.
	LOOPBACK:
	      ai3+   Line3 (PulseBlaster bit 7), 1k series resistor, 10k pullup.
	      ai2+   Op-amp buffered CY7-SD7 temperature sensor diode, 10k pullup.
	The connection between the ADC's inverting inputs (ai3-,ai2-) and VRst/2 is only made
	within the dewar; if it is detached, operate the NI4462 in pseudo-differential mode.

	The upper 3 bits of the ftdi245 device are used to control the clock division used for
	triggering the Pulseblaster from the external 5kHz clock: variable divisors can be
	selected, or it can be set to manual (external pulses, or pb_start).

OPTIONS:
	-f  FILTER_NAME | FILTER_NUMBER
	        set filter. The filter may be selected by name or by number. (See -l).
	        This causes the stepper motor to advance as many steps as necessary, until
	        the desired filter is reached.

	-p      get filter position. Read from the hardware and print it, e.g.:
	        "FILTER_NAME\\nCurrent filter is: #FILTER_NUMBER, FILTER_NAME (DESCRIPTION)"
	        Exit status 1 if misaligned.

	-l      list available filters, with a description of each of them.

	-c   CLOCK
	        set the clock divider to this value. (See -o for options).

	-d      get the clock divider setting (read back from device), and print it, e.g.:
	        "NAME\\nClock division is currently set to: NAME (DESCRIPTION)"

	-o      list the available clock divider options.

	-b      check battery level, and print it, e.g.:
	        "CHARGE_LEVEL\\nBattery level is: CHARGE_LEVEL (DESCRIPTION)"
	        Exit status 1 if flat.

	-m      monitor battery level. Polls every $INTERVAL_BATTERY_POLL_S seconds, and beeps if low. As this only
	        reads, not writes data, the process may be long-lived and co-exist with another instance.
	        [It yields temporarily, whenever the light-gates are being read.]

	-M      like m, but uses "\\r", polls every $INTERVAL_BATTERY_POLLFAST_S s, exits (1) when the battery goes flat. Eg:
	        $binary_name -M || zenity --error --title $binary_name --text='Flat Batt\\nCharge it.'

	-k      set up loopback mode. ai3,ai2 are swapped from the Hawaii quadrants (Q3,Q2), to
	        the secondary channels, (Line3, temperature diode) respectively, as described
	        above. This is useful for sending known data into the analog input, e.g. to
	        calibrate the FIFO delay-line for the NI 4462 trigger, and for measuring the
	        precise temperature adjacent to the Hawaii chip.
	        The loopback relays are reset by subsequent invocation of -z.
	        N.B. Do not operate camera in this state: data on Q3/Q2 will be invalid.

	-z      make the control circuit power down the dewar but do nothing else. This also
	        disconnects the loopback mode (-k). This is useful for initialisation after
	        system power-on.
	        [It is implicit in all instructions after they terminate, except loopback (-k).]

	-a      align. Constantly poll light-gate status, and print filter names. The stepper
	        motor is not driven. This is useful for manually aligning the filterwheel, when
	        assembling the dewar. See also -A and -S.

	-A      like -a, but print the raw data for each bit.

	-s      self-test. Run the filter wheel in constant rotation, listing the filter names
	        as they pass. Automatically stop after $SELFTEST_STEPS steps.

	-S      like -S, but wait for [ENTER] before proceeding to next step.

	-h      display this help message, and exit.

	-v      print verbose debug information (to stderr). Otherwise, be quiet.

	-w      print the configuration of this program, and exit.

	-i      ignore fatal errors: warn; don't quit.
	        Useful for flat-battery: some power may remain when 'running on empty'.

EXAMPLES:
	$binary_name -p                 #Print the current filter position
	$binary_name -f Green           #Select the filter named "Green"

STREAMS:
	All messages go to stderr, except the result of -p, -d, -b, -m which go to stdout.

EXIT CODES:
	On success, $binary_name will exit with status 0. If there is an error, the program
	will print a helpful message, and the exit with status 1.

DOCUMENTATION:
	See also filterctl(1), ftdi245io(1)

AUTHOR:
	$AUTHOR $EMAIL $COPYRIGHT_DATES
	Download from: $WEB
	$LICENSE

VERSION:
	Version $VERSION, released on $RELEASE_DATE.

EOT;


//--------------------------------------------------------------------------------------------------------------
//FUNCTIONS:

function output($msg){				//Output to STDOUT.
 	fwrite(STDOUT,"$msg\n");
}
function print_msg($msg){			//Print message to STDERR
	fwrite(STDERR,"$msg\n");
}
function print_word($word){			//Print word to STDERR, no newline
	fwrite(STDERR,"$word");
}
function print_status($msg){			//Print message to STDERR, followed by \r rather than \n, for status-updates.
	fwrite(STDERR, str_pad($msg,150) ."\r"); //Append some spaces, so as to ensure over-writing of all the previous content.
}
function fatal_error($msg,$exit_code=1){	//Print fatal error to STDERR and exit with error-code. (unless IGNORE_FATAL_ERRORS is true);
	global $IGNORE_FATAL_ERRORS;		//If $exit_code is not specified, by default, exit (1).
	static $count=0;
	if ($IGNORE_FATAL_ERRORS){
		fwrite(STDERR,"Fatal Error (Ignored): $msg\n");
	}else{
		fwrite(STDERR,"Fatal Error: $msg\n");
			$count++;
			if ($count >= 3){	//Note that dewar power off could itself fail with fatal error: don't get stuck!
				fwrite(STDERR,"Error handling fatal error (could not even turn off dewar power). Giving up.\n");
				exit ($exit_code);
			}
		dewar_power("off");		//For safety, (attempt to) turn off dewar after fatal error.
		exit ($exit_code);
	}
}
function print_warning($msg){			//Print warning to STDERR
	fwrite(STDERR,"WARNING: $msg\n");
}
function debug_print_msg($msg){			//Print message to STDERR, IFF DEBUG is true.
	global $DEBUG;				//NB: don't prepend anything.
	if ($DEBUG){
		fwrite(STDERR,"$msg\n");
	}
}
function beep(){				//ASCII Bell character. (Note some terminals break this because they think the user doesn't want to hear it...)
	fwrite(STDERR, chr(0x07));
}

function list_filters(){			//List the available filters.
	global $FILTERS;
	print_msg("The following filters are available. They may be selected (using -f) by either number or by name:\n\n".
		  "  NUMBER     NAME                 DESCRIPTION\n".
		  "  ------     ----                 -----------");
	for ($i=0; $i < count($FILTERS); $i++){
		print_msg("     ".str_pad($i,2)."      ".str_pad($FILTERS[$i]['key'],20)." ".$FILTERS[$i]['desc']);
	}
	print_msg("");
}

function list_clocks(){			//List the available filters.
	global $CLOCKS;
	print_msg("The following clock divisons are available. They may be selected (using -c) by name:\n\n".
		  "  NAME       DESCRIPTION\n".
		  "  ----       -----------");
	for ($i=0; $i < count($CLOCKS); $i++){
		print_msg("  ".str_pad($CLOCKS[$i]['key'],10)." ".$CLOCKS[$i]['desc']);
	}
	print_msg("");
}

function print_config(){			//Print configuration of this program (and list the filters). Some trivial things are omitted.
	global $binary_name, $AUTHOR, $EMAIL, $LICENSE, $VERSION, $RELEASE_DATE;
	global $DEBUG, $DUMMY, $FTDI245IO_BIN, $FTDI245IO_WRITE, $FTDI245IO_READBACK;  //[$DUMMY_BIN is not needed: it's already over-written $FTDI245IO_BIN if appropriate.]
	global $IO_MASK, $POWER_BIT, $LIGHT_GATE_BITS, $STEP_BIT, $CLOCK_DIVIDER_BITS;
	global $FILTERS, $VOLTAGES, $VOLTAGES_KEYS_ILLEGAL, $CLOCKS;
	global $INTERVAL_POSTPOWER_MS, $INTERVAL_PREPOWER_NOGLITCH_MS, $INTERVAL_STEP_MS, $INTERVAL_BATTERY_POLL_S, $INTERVAL_BATTERY_POLLFAST_S;
	print_msg("This is the configuration for $binary_name:");
	print_msg("   * Debugging is ".(($DEBUG)? ('enabled; stderr will be verbose.') : ('disabled; stderr will be terse.')));
	print_msg("\n   * ".(($DUMMY)? ("Dummy I/O hardware is on; FTDI245IO simulation program is '$FTDI245IO_BIN'") : ("Using real hardware I/O; FTDI program is $FTDI245IO_BIN (with args $FTDI245IO_WRITE or $FTDI245IO_READBACK).") ));
	print_msg("\n   * ".sprintf("The bit patterns are:\n\tIO_MASK:\t\t0x%02x\n\tPOWER_BIT:\t\t0x%02x\n\tLIGHT_GATE_BITS:\t0x%02x\n\tSTEP_BIT:\t\t0x%02x\n\tCLOCK_DIVIDER_BITS:\t0x%02x", $IO_MASK, $POWER_BIT, $LIGHT_GATE_BITS, $STEP_BIT, $CLOCK_DIVIDER_BITS));
	print_msg("\n   * The intervals are:");
	print_msg("        At dewar switch-on, prepare motor control to prevent glitch:         $INTERVAL_PREPOWER_NOGLITCH_MS ms");
	print_msg("        After dewar switch-on, to allow the electronics to stabilise:        $INTERVAL_POSTPOWER_MS ms");
	print_msg("        Length of the motor's step-pulse and damping (555's RC + pause):     $INTERVAL_STEP_MS ms");
	print_msg("        Battery polling interval in monitor mode (-M mode):                  $INTERVAL_BATTERY_POLL_S  s  ($INTERVAL_BATTERY_POLLFAST_S s)");
	print_msg("\n   * The following voltage thresholds are configured:\n".
		  "        #     HEX      NAME                 DESCRIPTION\n".
		  "        -     ----     ----                 -----------");
	for ($i=0; $i < count($VOLTAGES); $i++){
		if (!in_array($i, $VOLTAGES_KEYS_ILLEGAL)){
			print_msg("        ".str_pad($i,2)."    ". sprintf("0x%02x",$VOLTAGES[$i]['bits'])."     ".str_pad($VOLTAGES[$i]['key'],20). " ".$VOLTAGES[$i]['desc']);
		}
	}
	print_msg("\n   * The following filters are available:  (Wheel has $STEPPER_STEPS steps per revolution).\n".
		  "        #     HEX      NAME            SEQ  DESCRIPTION\n".
		  "        -     ----     ----            ---  -----------");
	for ($i=0; $i < count($FILTERS); $i++){
		print_msg("        ".str_pad($i,2)."    ". sprintf("0x%02x",$FILTERS[$i]['bits'])."     ".str_pad($FILTERS[$i]['key'],15). " ".str_pad($FILTERS[$i]['seq'],3). " ".$FILTERS[$i]['desc']);
	}
	print_msg("\n   * The following clock divisions are available:\n".
		  "        #     HEX      NAME                 DESCRIPTION\n".
		  "        -     ----     ----                 -----------");
	for ($i=0; $i < count($CLOCKS); $i++){
		print_msg("        ".str_pad($i,2)."    ". sprintf("0x%02x",$CLOCKS[$i]['bits'])."     ".str_pad($CLOCKS[$i]['key'],20). " ".$CLOCKS[$i]['desc']);
	}
	print_msg("\n    * Version $VERSION, on $RELEASE_DATE, written by $AUTHOR $EMAIL.");
	print_msg("      $LICENSE\n");
}

function ms_sleep($ms){
	debug_print_msg("ms_sleep: now sleeping for $ms ms...");
	usleep ($ms * 1000);
}


//Actually execute the command, and return the result string
function do_the_cmd($cmd){
	global $DEBUG;
	global $FTDI245IO_BIN;

	if ($DEBUG){
		$stderr_devnull='';		//Stderr gets passed through.
	}else{
		$stderr_devnull='2>/dev/null'; 	//Stderr discarded.
	}
	$cmd = trim ("$cmd $stderr_devnull");
	debug_print_msg("    About to run command: $cmd");	//Run the command. Return value is in $retval, $output_array gets stdout (one line per element). $result is the last line of the output.
	$result=exec($cmd, $output_array, $retval);		//NB STDERR is passed directly through to this script's stderr (unless redirected by $stderr_devnull).

	if ($retval == 127){					//Return status? 127 means not found, 0 means success.
		fatal_error("could not run command '$cmd'. Is the program '$FTDI245IO_BIN' installed and in the path?");
	}else if ($retval != 0){
		fatal_error("command '$cmd' failed, exit status $retval.");  //See ../../ftdi245io/src/ftdi245io.h for definitions
	}else{							//$command executed OK.
		if ($DEBUG){
			if (count($output_array) <= 1){
				print_msg("    The result of the command was: $output_array[0]");
			}else{
				print_msg("    The result of the command was:");
				foreach ($output_array as $line){
					print_msg("      $line");
				}
			}
		}
	}
	return (trim($result));
}

//Read ftdi245; Get the bits specified by mask.  (Mask is 1 for the bits we want; 0 for the ones we don't).
function get_bits ($mask){
	global $FTDI245IO_BIN, $FTDI245IO_READBACK;
	debug_print_msg( sprintf ("get_bits(0x%02x)", $mask));
	$cmd = "$FTDI245IO_BIN $FTDI245IO_READBACK"; 	//Eg  ftdi245io --porta_readback.   Read back data from the port, including values at the FTDI's buffer for output pins.
	$result = do_the_cmd ($cmd);
	$byte = hexdec ($result);			//Input should be as single hex-representation of a byte in the form "0x34". It's a *string* so parse using hexdec()
	$bits = $byte & $mask;
	debug_print_msg( sprintf("Read data byte from interface: 0x%02x, which becomes 0x%02x when masked. (Used mask 0x%02x.)", $byte, $bits, $mask) );
	return ($bits);
}

//Modify the bits specified by mask. Read the byte from ftdi245 hardware. THEN update (set or clear) the bits (subject to mask), then write it back.
//Note: we are modifying the persistent state stored in the hardware's output buffer. Useful if we don't want to change the CLOCK outputs.
function set_bits ($mask, $bits){	//$mask is what we want to change; $bits are the values we want it/them to become. Eg to set pin_x, call this as  'set_bits(pin_x, 0xff)'.
	global $FTDI245IO_BIN, $FTDI245IO_READBACK, $FTDI245IO_WRITE, $IO_MASK;
	if ($bits === 'HIGH'){		//Accept HIGH and LOW as shortcuts for 0xff and 0x00.
		$bits = 0xff;		//Accept !x as the inverse.
	}elseif ($bits === 'LOW'){
		$bits = 0x00;
	}elseif ($bits === '!HIGH'){
		$bits = 0x00;
	}elseif ($bits === '!LOW'){
		$bits = 0xff;
	}
	debug_print_msg( sprintf ("set_bits(0x%02x, 0x%02x)", $mask, $bits));
	if ($mask & ~$IO_MASK){
		fatal_error (sprintf("requested mask 0x%02x is not within output mask 0x%02x.", $mask, $IO_MASK ));  //Sanity check: if we are trying to alter bits we shouldn't. IO_MASK considers 1s as outputs.
	}
	$cmd = "$FTDI245IO_BIN $FTDI245IO_READBACK";   //Get the current value of the output buffer.
	$result = do_the_cmd ($cmd);
	$byte = hexdec ($result);

	$setme = $bits & $mask;				//Modify the bits we specify in $mask. Leave the rest unaltered.
	$clearme = $mask & (~$bits);
	$newbyte = ($byte | $setme) & ~$clearme;
	$newbyte_str = sprintf("0x%02x", $newbyte);

	$cmd = "$FTDI245IO_BIN $FTDI245IO_WRITE $newbyte_str"; //Write back the updated value.
	$result = do_the_cmd ($cmd);
	//$result is not used. 				//Result is a one line summary, eg 'Written byte 0x40 to PortA (raw: 0x44, iomask: 0xF8).'

	debug_print_msg( sprintf("Modified interface byte from 0x%02x to 0x%02x. (Used mask 0x%02x with bits 0x%02x: set bits 0x%02x, clear bits 0x%02x.)", $byte, $newbyte, $mask, $bits, $setme, $clearme) );
	return ($bits);
}

function dewar_power($onoff){	//Power up the contents of the dewar. $onoff is either 'on' or 'off'. Before turning on, pre-set the step_bit, so that the motor doesn't glitch.
	global $POWER_BIT, $POWER_BIT_ACTIVE, $STEP_BIT, $STEP_BIT_ACTIVE_GOING;
	global $INTERVAL_POSTPOWER_MS, $INTERVAL_STEP_MS, $INTERVAL_PREPOWER_NOGLITCH_MS;

	$orig=get_dewar_power();

	$step_bit = get_bits ($STEP_BIT);		//If the step bit is already in the active state, then possibly the 555 had just been triggered. Sleep briefly to let it time-out.
	if (( ($STEP_BIT_ACTIVE_GOING == "HIGH") and ($step_bit) ) OR ( ($STEP_BIT_ACTIVE_GOING == "LOW") and (!$step_bit) )){
		debug_print_msg ("Step bit was active; sleeping $INTERVAL_STEP_MS ms to allow 555 to time out.");
		ms_sleep ($INTERVAL_STEP_MS);
	}

	if ($onoff == 'on'){				//Turn on power:
		set_bits ($STEP_BIT, '!'.$STEP_BIT_ACTIVE_GOING); //prevent motor from kicking: take 555's trigger bit inactive. before applying power...
		ms_sleep($INTERVAL_PREPOWER_NOGLITCH_MS); 	  //wait for stabilisation. (probably unnecessary now that we explicitly check the step-bit's state above).
		set_bits ($POWER_BIT, $POWER_BIT_ACTIVE);	  //set the power bit
		ms_sleep($INTERVAL_POSTPOWER_MS);		  //wait for stabilisation.  Not too short, else after power-on, we can obtain wrong values from the sensors.

	}elseif ($onoff == 'off'){			//Turn off power:
		set_bits ($POWER_BIT, '!'.$POWER_BIT_ACTIVE);	  //clear the power bit
		set_bits ($STEP_BIT, '!'.$STEP_BIT_ACTIVE_GOING); //Leave the STEP bit in the inactive state for next time. Disconnect the loopback.
		if ($orig != 'off'){				  //Skip unneeded sleep if we were already off.
			ms_sleep($INTERVAL_POSTPOWER_MS);	  //wait for stabilisation.  Not too short, else after power-on, we can obtain wrong values from the sensors.
		}
	}else{
		fatal_error("dewar_power wrong argument: '$onoff'");
	}
	debug_print_msg("Dewar power has been turned $onoff.");
}

function get_dewar_power(){	//Return 'on' or 'off', depending on dewar power state. This just checks the value, but doesn't change it.
	global $POWER_BIT, $POWER_BIT_ACTIVE;
	$dewar_bit = get_bits($POWER_BIT);
	if (( ($POWER_BIT_ACTIVE == "HIGH") and ($dewar_bit) ) OR ( ($POWER_BIT_ACTIVE == "LOW") and (!$dewar_bit) )){
		debug_print_msg ("Dewar power was on.");
		return "on";
	}else{
		debug_print_msg ("Dewar power was off.");
		return "off";
	}
}


function connect_loopback(){	//Connect the loopback. This is done by having STEP bit ACTIVE (the state, not the edge), combined with dewar power off.
	global $STEP_BIT, $STEP_BIT_ACTIVE_GOING;	//[It's a late-addition to the hardware, and utilises an otherwise unused condition to control it over the existing data lines.]
	dewar_power("off");				//Power off first, so that we don't step the motor by accident! (and, for that matter, implicitly unset $STEP_BIT)
	set_bits ($STEP_BIT, $STEP_BIT_ACTIVE_GOING);
}

function read_filter_position(){ //Read the current filter position from the hardware, and look it up in $FILTERS. Return the relevant key.
	global $LIGHT_GATE_BITS;
	global $FILTERS;

	$light_gate_key = get_bits ($LIGHT_GATE_BITS);  //Read the values of the bits we care about.

	$filter_number=false;	//Look this up in $FILTERS.
	for ($i=0; $i< count ($FILTERS); $i++) {
		if ($FILTERS[$i]['bits'] == $light_gate_key){
			$filter_number=$i;
			break;
		}
	}
	if ($filter_number===false){
		fatal_error(sprintf("error looking up filter. Searching for 0x%02x  (Input is 0x%02x, light_gate mask is 0x%02x)", $light_gate_key, $light_gate_key, $LIGHT_GATE_BITS));
	}
	return ($filter_number);
}


function step_motor(){					//Step around once. Check for feedback. Wait briefly after the step has been completed.
	global $STEP_BIT, $STEP_BIT_ACTIVE_GOING;	//[NOTE: any error messages need a leading \n, because we are probably embedded within "step................... [...]"
	global $INTERVAL_STEP_MS;
	#Note: the original circuit actually checked using feedback whether the motor was pulsing. Useful for verification, and if we didn't know how long the pulse should be.
	#Now the lines are used for other things, and there isn't one spare. So the 555 monostable's timeout must be second-guessed. OK to be too large, try not to be too small.

	set_bits ($STEP_BIT, $STEP_BIT_ACTIVE_GOING);	//Rising edge on step bit. Trigger the step.
	ms_sleep($INTERVAL_STEP_MS );			//Wait for the length of the step. (second-guess the 555 monostable). Then let the mechanics stabilise and the 555 prepare for next pulse.
	set_bits ($STEP_BIT, '!'.$STEP_BIT_ACTIVE_GOING);//Falling edge on step bit. Ready for next step.
}

function read_clock_state(){ 				//Read the current clock division state from the hardware, and look it up in $CLOCKS. Return the relevant key.
	global $CLOCK_DIVIDER_BITS;
	global $CLOCKS;

	$clock_key = get_bits ($CLOCK_DIVIDER_BITS);  	//Read the values of the bits we care about.

	$clock_number=false;	//Look this up in $CLOCKS.
	for ($i=0; $i< count ($CLOCKS); $i++) {
		if ($CLOCKS[$i]['bits'] == $clock_key){
			$clock_number=$i;
			break;
		}
	}
	if ($clock_number===false){
		fatal_error(sprintf("error looking up clock value. Searching for 0x%02x  (Input is 0x%02x, clock_divider mask is 0x%02x)", $clock_key, $clock_key, $CLOCK_DIVIDER_BITS));
	}
	return ($clock_number);
}

function read_battery_state($low_priority=false){		//Read battery state. Normally, WARN if the dewar wasn't already in battery mode as we expected, then turn it off and read the battery.
	global $LIGHT_GATE_BITS, $VOLTAGES, $VOLTAGES_KEYS_ILLEGAL;	//BUT, if $low_priority is set, then leave the dewar on, and return false.
	$dewar_power = get_dewar_power();	//returns "on" or "off",
	if ($dewar_power == 'on'){		//If it was on:
		if ($low_priority){		//	- if $low_priority, just return false.
			return (false);
		}				//	- normally, warn, turn it off, and continue.
		print_msg("WARNING: dewar was found in powered-on (rather than battery monitor) mode. Should this have happened?");
		dewar_power("off");
	}

	$voltage_key = get_bits($LIGHT_GATE_BITS);

	$voltage_number=false;	//Look this up in $VOLTAGES.
	for ($i=0; $i< count ($VOLTAGES); $i++) {
		if ($VOLTAGES[$i]['bits'] == $voltage_key){
			$voltage_number=$i;
			break;
		}
	}
	if ($voltage_number===false){
		fatal_error(sprintf("error looking up voltage value. Searching for 0x%02x  (Input is 0x%02x, voltage_bits mask is 0x%02x)", $voltage_key, $voltage_key, $LIGHT_GATE_BITS));
	}
	if (in_array($voltage_number, $VOLTAGES_KEYS_ILLEGAL)){	//Is this an illegal value?
		fatal_error(sprintf("error: the voltage value detected is \"impossible\". Input of 0x%02x should never happen unless hardware is faulty. Check for broken fibre?", $voltage_key));
	}
	return ($voltage_number);
}

function check_battery_ok(){		//Return silently if battery OK; print error and exit (1) if flat. We don't want a bogus reading.
	global $VOLTAGES_KEY_FLAT;
	global $IGNORE_FATAL_ERRORS;
	if (read_battery_state() === $VOLTAGES_KEY_FLAT){
		if ($IGNORE_FATAL_ERRORS){ //Running on empty? Maybe.
			$flat_txt = " (Attempting to 'run on empty tank', requested by '-i'.)";
		}
		fatal_error ("Battery voltage is flat. Dewar data will be wrong.$flat_txt");
	}
}

function check_dewar_off(){		//Expect dewar to be off. If it is on, warn and turn it off!
	if (get_dewar_power() == 'on'){
		print_msg ("Warning: dewar power was on, unexpectedly. Is this an error?");
		dewar_power('off');
	}
}

function trap_ctrlc(){				//In some cases, we need to trap Ctrl-C, so that the dewar is turned off when we exit.
	declare(ticks = 1);
	if (!function_exists("sig_handler")){
		function sig_handler($signo) {
			dewar_power('off'); 	//Turn off the dewar. Then wait briefly, and exit.
			print_msg("\n\nKilled. Dewar power (motor/light-gates) is now off; loopback disconnected.");
			exit (0);
		}
	}
	if (function_exists("pcntl_signal")){	//In very old versions of PHP, this extension doesn't exist!
		pcntl_signal(SIGINT, "sig_handler");
		pcntl_signal(SIGTERM, "sig_handler");
		pcntl_signal(SIGQUIT, "sig_handler");
	}else{
		debug_print_msg("php-pcntl extension is missing; Ctrl-C will not be trapped.");
	}
}

function wait_for_keypress(){
        $read=array(STDIN);	//Wait for keypress, then continue.
        $write=NULL;            //Things to monitor. We only care about reading STDIN.
        $except=NULL;
	$timeout=NULL;          //(Blocking read).
        $num_changed_streams = @stream_select($read,$write,$except,$timeout);   //Stream select: returns the number of streams that wouldn't block, if they were now read. [Supress warning on Ctrl-C, with '@']
        if ($num_changed_streams===false){      //Error. Shouldn't happen. Ignore this.
		print_notice("stream_select() returned false. Ignoring it.");
                $line=false;
	}elseif ($num_changed_streams > 0){     //At least one stream (actually, exactly one in this case) changed status.
		$line=strtolower(trim(fgets(STDIN)));   //Read line from stdin
	}else{
		$line=false;                    //No stream changed status. Nothing to do. (shouldn't happen in blocking mode).
        }
        return($line);   //Be careful with return value (false vs. '')
}

//--------------------------------------------------------------------------------------------------------------
// GET COMMAND-LINE ARGUMENTS:

$flags="aAbdhiklmMopsSvwzc:f:";		//Each letter listed here is a possible flag. Letters followed by colon may take an argument.
$options_array=getopt($flags);  	// '-h' '--h' '-f filter_name' are all acceptable. [This is similar to GNU getopts]

function bug_check($key,$value){	//For some daft reason, "-i -o foo" is parsed as "$i=-o; foo" , NOT as "$i=; $o=foo"
	if (substr($value,0,1)=='-'){	//So, it is possible that a *missing* option will appear to be the next flag instead!
		fatal_error("option '$key' cannot take value '$value' which begins with '-'");  //Check that $value does not begin with '-'
	}
}

foreach ($options_array as $key => $value){
	switch ($key){
		case 'a':					//alignment help: constantly poll and print filter wheel status.
			$do_align_check=true;
			break;

		case 'A':					//alignment help: in a "raw" mode.
			$do_align_raw_check=true;
			break;

		case 'b':					//check and print battery level.
			$do_battery_check=true;
			break;

		case 'c':					//set clock divider to this value
			$set_clock=true;
			$requested_clock=trim($value);
			bug_check($key,$value);
			break;

		case 'd':					//get and print the clock divider setting.
			$read_clock=true;
			break;

		case 'f':					//goto filter.
			$goto_filter=true;
			$requested_filter=trim($value);
			bug_check($key,$value);
			break;

		case 'h':					//help   [Also works with -help, --help]
			print_msg($USAGE);
			exit (0);
			break;

		case 'i':					//attempt to ignore fatal errors (eg run on "empty" with flat battery)
			$IGNORE_FATAL_ERRORS=true;
			break;


		case 'k':					//enable loopback mode.
			$do_connect_loopback=true;
			break;

		case 'l':					//list available filters
			list_filters();
			exit (0);
			break;

		case 'm':					//monitor battery level (keep polling).
			$do_battery_monitor=true;
			break;

		case 'M':					//monitor battery level (keep polling); exit on flat.
			$do_battery_monitor=true;
			$battery_monitor_exit_when_flat=true;
			$battery_monitor_poll_faster=true;
			$battery_monitor_use_r=true;
			break;

		case 'o':					//list clock options.
			list_clocks();
			exit (0);
			break;

		case 'p':					//read current position and print it.
			$read_filter_position=true;
			break;

		case 's':					//do active selftest: constantly step.
			$do_selftest=true;
			break;

		case 'S':					//do manual selftest: interactively step.
			$do_selftest=true;
			$selftest_manual_step=true;
			break;

		case 'v':					//verbose/debug
			$DEBUG=true;
			print_msg("Debugging enabled...");
			break;

		case 'w':					//print config
			print_config();
			exit (0);
			break;

		case 'z':					//power down the dewar.
			$power_down=true;
			break;

		default:
			print_msg("Invalid argument: $key. Use -h for help.");
			exit (1);
	}
}

if ($argc == 1 or count($options_array) == 0){  //Wrong args, or wrong options (eg -f has been supplied without a value.)
	print_msg("Wrong arguments. Use -h for help.");
	exit (1);
}

//--------------------------------------------------------------------------------------------------------------
//MAIN PROGRAM. Here, we have to deal with either: $goto_filter($requested_filter), $read_filter_position, $set_clock($requested_clock), $read_clock, $power_down, $do_battery_check, $do_battery_monitor, $do_align_check, $do_align_raw_check, or $do_selftest.

trap_ctrlc();   //Make sure we trap Ctrl-C, and turn off the dewar before exiting.

//Read filter position and print it.
if ($read_filter_position){
	check_battery_ok();	//Check battery is ok, so we can trust the readings.
	dewar_power('on');	//Turn on the dewar sensor. Then wait briefly for it to settle.

	$filter_number=read_filter_position(); //Read the current filter position from the hardware, and look it up in $FILTERS.

	//Print some information to STDOUT, in the format: "FILTER_NAME\nCurrent filter is: #X, FILTER_NAME (FILTER_DESCRIPTION)
	output($FILTERS[$filter_number]['key']."\nCurrent filter is: #$filter_number, ".$FILTERS[$filter_number]['key']." (".$FILTERS[$filter_number]['desc'].")");

	dewar_power('off'); 	//Turn off the dewar. Then wait briefly, and exit.
	print_msg("Dewar power (motor/light-gates) is now off; loopback disconnected.");
	check_battery_ok();	//Still ok?

	if ( ($filter_number == $FILTER_MISALIGNED) or ($filter_number == $FILTER_ILLEGAL) ){  //Exit 1 if misaligned (useful for other scripts)
		fatal_error ("Filter position is misaligned.");
	}
}


//Seek to a specified filter.
if ($goto_filter){
	check_battery_ok();	//Check battery is ok, so we can trust the readings.

	dewar_power('on');	//Turn on the dewar sensor. Then wait briefly for it to settle.

	if ( ($requested_filter=='') and ($requested_filter!==0) ){  	//Was a filter specified? Match '', except when === 0
		fatal_error("missing argument: no target filter was specified.");
	}else if (is_numeric($requested_filter)){			//Look up the requested filter. Specified either by an integer, or a string.
		if (array_key_exists($requested_filter,$FILTERS)){	//Match integer
			$target_filter_number=$requested_filter;
			$target_filter_name=$FILTERS[$requested_filter]['key'];
		}else{
			fatal_error("Unknown target filter '$requested_filter'. Please specify either by number or by name; use -l for a list.");
		}
	}else{
		$target_filter_number=false;				//Match string (case-insensitive match)
		for ($i=0; $i < count ($FILTERS); $i++){
			if ( trim(strtolower($FILTERS[$i]['key'])) == trim(strtolower($requested_filter)) ){
				$target_filter_number=$i;
				$target_filter_name=$FILTERS[$i]['key'];
				break;
			}
		}
		if ($target_filter_number === false){
			fatal_error("Unknown target filter '$requested_filter'. Please specify either by number or by name; use -l for a list.");
		}
	}
	debug_print_msg("Trying to step to filter #$target_filter_number ($target_filter_name)...");

	if ($target_filter_number == $FILTER_MISALIGNED){	//Check for idiot requests.
		fatal_error("You requested to step to filter #$target_filter_number ($target_filter_name). That is a misaligned position, so it would be silly."); //[OK, technically, we could pick *any* misaligned spot, but it's silly.]
	}elseif ($target_filter_number == $FILTER_ILLEGAL){
		fatal_error("You requested to step to filter #$target_filter_number ($target_filter_name). That is an illegal position, so is not possible."); //Why *ask* for a hardware error?
	}

	$filter_number=read_filter_position(); 	//Where are we to begin with?
	$steps_so_far=0;			//Keep stepping till we get there....
	while ($filter_number != $target_filter_number){
		print_word("Step:$steps_so_far->");
		step_motor(); 			//step once.
		$steps_so_far++;
		$filter_number=read_filter_position(); 	//Where are we now?
		print_word("#$filter_number; ");
		if ($steps_so_far >= $MAX_STEP_ATTEMPTS){ //Give up eventually.
			print_msg("\n");
			fatal_error("We've stepped $steps_so_far times now, and still haven't reached alignment. Something's wrong; giving up.");
		}
	}

	if ($steps_so_far==0){		//We have arrived, after $steps_so_far steps.
		print_msg("We were already at our destination (Filter #$target_filter_number, $target_filter_name). No steps were required.");
	}else{
		print_msg("\nWe've reached our destination (Filter #$target_filter_number, $target_filter_name), after $steps_so_far steps.");
	}
	
	dewar_power('off'); 	//Turn off the dewar. Then wait briefly, and exit.
	print_msg("Dewar power (motor/light-gates) is now off; loopback disconnected.");
	check_battery_ok();	//Still ok?
}



//Alignment check: constantly poll filter-wheel position, and print it. useful when aligning the dewar mechanics.
if ($do_align_check){

	print_msg("Alignment check. Will read and print filter status every $INTERVAL_ALIGNMENTCHECK_MS ms...");
	print_msg("The motor will not be stepped; move the wheel manually. Ctrl-C to quit.");

	check_battery_ok();	//Check battery is ok, so we can trust the readings.
	dewar_power('on'); 	//Turn on the dewar. Then wait briefly to settle.

	print_msg ("\nCOUNT    READ_BYTE    FILTER_NUMBER     FILTER_NAME");
	print_msg (  "-----    ---------    -------------     -----------");
	$i=0;
	while (true){
 		$filter_number=read_filter_position();		//Read the current filter position.
		$bits = get_bits ($LIGHT_GATE_BITS);	 	//Get raw bits (requires another read, or a re-factor of read_filter_position() )
 		print_status(" ".str_pad($i,6)."  ".sprintf("0x%02x",$bits)."         ".str_pad("#$filter_number",17)." ".$FILTERS[$filter_number]['desc']);	//N.B. Followed by \r, rather than \n
		ms_sleep ($INTERVAL_ALIGNMENTCHECK_MS);
		$i++;
	}
	//Infinite loop. We never get here, so no chance to power off the dewar contents. Hence we need to trap Ctrl-C. Else, remember to -z manually.
}
//Alignment check: raw data. Useful to calibrate the phototransistor bias.
if ($do_align_raw_check){

	print_msg("Alignment check (raw). Will read and print light-gate status $INTERVAL_ALIGNMENTCHECK_MS ms...");
	print_msg("The motor will not be stepped. High means that the gate is open.  Ctrl-C to quit. ");

	check_battery_ok();	//Check battery is ok, so we can trust the readings.
	dewar_power('on'); 	//Turn on the dewar. Then wait briefly to settle.

	//$LIGHT_GATE_BITS: correspond to the 3 light gates. (Bits 2,1,0 are the three sensors; 8 permutations. High means light gate is open).
	print_msg ("\nCOUNT    BIT_2      BIT_1      BIT_0      FILTER");
	print_msg (  "-----    --------   --------   --------   ------");
	$i=0;
	while (true){
 		$filter_number=read_filter_position();		//Read the current filter position.
		$bits = get_bits ($LIGHT_GATE_BITS);		//Get bits. As above, this duplicates a hardware read for simplicity, should probably re-factor.
		$bit2 = ($bits & 0x04) ? "1:open" : "0:closed";
		$bit1 = ($bits & 0x02) ? "1:open" : "0:closed";
		$bit0 = ($bits & 0x01) ? "1:open" : "0:closed";
 		print_status(" ".str_pad($i,6)."  ".str_pad($bit2,8)."   ".str_pad($bit1,8)."   ".str_pad($bit0,8)."   ". $FILTERS[$filter_number]['key']);	//N.B. Followed by \r, rather than \n
		ms_sleep ($INTERVAL_ALIGNMENTCHECK_MS);
		$i++;
	}
	//Infinite loop. We never get here, so no chance to power off the dewar contents. Hence we need to trap Ctrl-C. Else, remember to -z manually.
}

//Do active selftest - step the motor repeatedly, and print the filter statuses each time.
//This tests the entire chain of hardware: ftdi245io, the usb device, the filter_pcb, the psu, the motor, optics, and the mechanics.
if ($do_selftest){
	check_battery_ok();	//Check battery is ok, so we can trust the readings.
	dewar_power('on'); 	//Turn on the dewar. Then wait briefly to settle.
	print_msg("Dewar power has been turned on.");

	print_msg ("Self-test mode. Will step $SELFTEST_STEPS times, or Ctrl-C to quit...");
	if ($selftest_manual_step){
		print_msg ("Press [ENTER] after each step to continue.");
	}
	print_msg ("\nSTEP    FILTER_NUMBER     FILTER_NAME");
	print_msg ("----    -------------     -----------");

	$consecutive_misalignments=0;
	$filter_number=read_filter_position();		//Read the initial filter position.
	$last_real_filter_number=$filter_number;

	for ($i=0; $i < $SELFTEST_STEPS; $i++){		//NB not too many, or wear on the bearings will result - the motor life is not that large. Don't leave it running for a week!
		step_motor();				//Step around once.
		$filter_number=read_filter_position();	//Read the current filter position.
		$filter_seq = $FILTERS[$filter_number]['seq'];   //sequence order in physical wheel.
 		print_msg(" ".str_pad($i,6)." ".str_pad("#$filter_number",17)." ".$FILTERS[$filter_number]['key']);

		if ($filter_number == $FILTER_MISALIGNED){ //What if there are too many misalignments in a row?
			$consecutive_misalignments++;
			if ($consecutive_misalignments > $SELFTEST_MAX_MISALIGNED){  //More than $SELFTEST_MAX_MISALIGNED misalignments in a row?
				print_msg("Warning: there have now been $consecutive_misalignments consecutive steps resulting in a misalignment.");
				$consecutive_misalignments=0;
			}
		}elseif ($filter_number == $FILTER_ILLEGAL){  //Illegal result?
			print_msg("Warning: selftest has reached illegal position: lightgates should never all be on simultaneously.");
		}else{					//We are on an actual filter.
			if ( ($last_real_filter_seq) and  ($filter_seq - $last_real_filter_seq !=1) and ($last_real_filter_seq - $filter_seq != count($FILTERS) - 3) ){ //Check that we have moved by one, in the right direction.
				print_msg("Warning: last step was not correctly sequential (Wheel went from sequence #$last_real_filter_seq to #$filter_seq)."); //1->2->3->4->5->6->1. for SEQUENCE, not filter number!
			}
			$last_real_filter_seq=$filter_seq;
			$consecutive_misalignments=0;	//reset counter
		}
		if ($selftest_manual_step){
			print_word ("Next?: ");
			wait_for_keypress();
		}
	}
	print_msg("Selftest completed after $SELFTEST_STEPS steps. Now stopping, to save motor (which, being LN2/vacuum-rated is not lubricated, and has a relatively short life).");

	dewar_power('off'); 	//Turn off the dewar. Then wait briefly, and exit.
	print_msg("Dewar power (motor/light-gates) is now off; loopback disconnected.");
	check_battery_ok();	//Check battery is ok, so we can trust the readings.
}

//Read clock divider setting from hardware and print it.
if ($read_clock){
	$clock_number = read_clock_state(); //Read the clock number back from the hardware, and look it up in $CLOCKS.

	//Print some information to STDOUT, in the format: NAME\nClock division is currently set to:...'
	output($CLOCKS[$clock_number]['key']."\nClock division is currently set to: ".$CLOCKS[$clock_number]['key']." (".$CLOCKS[$clock_number]['desc'].")");

	check_dewar_off(); 	//Redundant, hopefully!
}


//Set a specified clock ($requested_clock)
if ($set_clock){
	$target_clock_number=false;			//Match string (case-insensitive match)
	for ($i=0; $i < count ($CLOCKS); $i++){
		if ( trim(strtolower($CLOCKS[$i]['key'])) == trim(strtolower($requested_clock)) ){
			$target_clock_number=$i;
			$target_clock_name=$CLOCKS[$i]['key'];
			$target_clock_desc=$CLOCKS[$i]['desc'];
			$target_clock_bits=$CLOCKS[$i]['bits'];
			break;
		}
	}
	if ($target_clock_number === false){
		fatal_error("Unknown target clock '$requested_clock'. Please specify by frequency; use -o for a list.");
	}
	debug_print_msg("Setting clock to  #$target_clock_number ($target_clock_name)...");

	set_bits($CLOCK_DIVIDER_BITS, $target_clock_bits);	//Actually do it.
	print_msg("Clock division now set to: $target_clock_name,  $target_clock_desc.");

	check_dewar_off();	 //Redundant, hopefully!
}

if ($do_battery_check){
	$voltage_number = read_battery_state(); //Read the voltage number back from the hardware, and look it up in $VOLTAGES.

	//Print some information to STDOUT, in the format: "CHARGE_LEVEL\n\nBattery level is: CHARGE_LEVEL (DESCRIPTION)"
	output($VOLTAGES[$voltage_number]['key']."\nBattery Level is: ".$VOLTAGES[$voltage_number]['key']." (".$VOLTAGES[$voltage_number]['desc'].")");

	//If it was flat, print warning and exit 1.  Note that 'flat' is calibrated as slightly above complete circuit power-down, so there is a slight grace-period.
	if ($voltage_number == $VOLTAGES_KEY_FLAT){      //similar to a car fuel tank being "empty".
		fatal_error ("WARNING! Battery is flat. No response from dewar. It may, or may not, still have a small amount of power left to power the sensor; safety cutout is at $VOLTAGE_SAFETY_CUTOUT.");
	}

	check_dewar_off(); 	//Redundant, hopefully!
}

if ($do_battery_monitor){	//Always loop, every 60 seconds. Print the value. Every 20 lines, reprint the header. If flat, warn and beep.
	//This is intended to be a long-lived process, so we might be monitoring the battery while another process changes the filter. If so, we want to allow
	//that to happen. Call read_battery_state($low_priority=true), and rather than forcing off dewar power for our own needs, this will simply return false.
	//In -M mode, use "\r" rather than "\n", and poll every second.
	$i = 0;
	while (true){		//Read the voltage number back from the hardware, and look it up in $VOLTAGES.
		if ( ($i == 0) or ( ($i % 20 == 0) and (!$battery_monitor_use_r) ) ){
			output("\nMinutes    Status    Battery voltage details                     Notes\n-------    ------    ---------------------------------------     -----");
		}
		if ($battery_monitor_poll_faster){
			$min = floor ($i/60) . ":" . str_pad (($i%60),2,"0",STR_PAD_LEFT);
			$sleeptime = $INTERVAL_BATTERY_POLLFAST_S;
		}else{
			$min = $i;
			$sleeptime = $INTERVAL_BATTERY_POLL_S;
		}
		//Used to do read_battery_state(false) the first time, i.e. asserting that WE are in charge. Thats nice, but can possibly break another measurement, so something unpredictable could happen
		//Also, used to set $IGNORE_FATAL_ERRORS=true as a workaround for ftdi245io sometimes racing against itself. ftdi245io now uses flock(), so unnecessary.
		$voltage_number = read_battery_state(true);
		if ($voltage_number !== false){
			$flat_txt = '';
			if ($voltage_number === $VOLTAGES_KEY_FLAT){
				beep();
				$flat_txt = "!! WARNING !! No response from dewar. Sensor power failed or failing.";
			}
			$msg = " ".str_pad($min,6)."    ".str_pad($VOLTAGES[$voltage_number]['key'],8)."  Voltage: ".str_pad($VOLTAGES[$voltage_number]['desc'],34)." $flat_txt";
			if ($battery_monitor_use_r){
				print_status ($msg);
			}else{
				output ($msg);
			}
		}else{		//read_battery_state yielded to another process.
			$msg = " ".str_pad($min,6)."    Yield     Battery unreadable                          Another process is controlling the dewar; yielding to it.";
			if($battery_monitor_use_r) {
				print_status($msg);
			}else{
				output ($msg);
			}
		}
		if (($flat_txt) and ($battery_monitor_exit_when_flat)){ //exit on flat, if invoked as -M.
			output ("\n"); //important with $battery_monitor_use_r
			exit(1);
		}
		sleep ($sleeptime);
		$i++;
	}
	//never get here, unless Ctrl-C'd
}

//Connect loopback (from Line3 to Q3+ and temp-sensor to Q2+)
if ($do_connect_loopback){
	check_battery_ok();	//Ensure battery is OK; else we can't actually go into loopback mode (no power for the reed relay).
	connect_loopback();	//This slightly "misuses" the existing 2 data lines into the dewar to add a new feature. The relay swaps from Hawaii to LineClock3 if (step-bit-active AND dewar-power-off)
	debug_print_msg("Loopback connected (Q3+ to Line3, Q2+ to temperature, via SPDT reed relays.");
	print_msg("Loopback connected. To disconnect, use -z.");
}

//Power off.
if ($power_down){
	dewar_power('off');	//Turn off the dewar (or ensure it's already off). Then wait briefly, and exit.
	print_msg("Dewar power (motor/light-gates) is now off; loopback disconnected.");
}


exit (0);  //If we get to here, we're happy!
?>
